// Copyright (c) 2016 nyorain 
// Distributed under the Boost Software License, Version 1.0.
// See accompanying file LICENSE or copy at http://www.boost.org/LICENSE_1_0.txt

///\file
///\brief Makes the Vec specialization for reference types work.

#pragma once

#ifndef NYTL_INCLUDE_REFVEC_HPP
#define NYTL_INCLUDE_REFVEC_HPP

#include <nytl/vec.hpp>
#include <nytl/bits/refVecIterator.inl>

namespace nytl
{

template<size_t D, typename T> using RefVec = Vec<D, T&>;

template<typename T> using RefVec2 = Vec<2, T&>;
template<typename T> using RefVec3 = Vec<3, T&>;
template<typename T> using RefVec4 = Vec<4, T&>;

using RefVec2f = RefVec2<float>;
using RefVec2i = RefVec2<int>;
using RefVec2ui = RefVec2<unsigned int>;
using RefVec2d = RefVec2<double>;
using RefVec2b = RefVec2<bool>;

using RefVec3f = RefVec3<float>;
using RefVec3i = RefVec3<int>;
using RefVec3ui = RefVec3<unsigned int>;
using RefVec3d = RefVec3<double>;
using RefVec3b = RefVec3<bool>;

using RefVec4f = RefVec4<float>;
using RefVec4i = RefVec4<int>;
using RefVec4ui = RefVec4<unsigned int>;
using RefVec4d = RefVec4<double>;
using RefVec4b = RefVec4<bool>;

template<size_t D, typename T>
class Vec<D, T&>
{
public:
	static constexpr std::size_t dim = D;

	using Size = std::size_t;
	using Precision = T;
	using Value = T;
	using VecType = Vec<dim, T>;
	using RefVecType = Vec<D, T&>;

	using Reference = Value&;
	using ConstReference = const Value&;
	using Pointer = Value*;
	using ConstPointer = const Value*;
	using Iterator = RefVecIterator<dim, Value>;
	using ConstIterator = ConstRefVecIterator<dim, Value>;
	using ReverseIterator = std::reverse_iterator<Iterator>;
	using ConstReverseIterator = std::reverse_iterator<ConstIterator>;
	using Difference = Size;

	using RangeError = std::out_of_range;

	//stl
	using size_type = Size;
	using value_type = Value;
	using reference = Reference;
	using const_reference = ConstReference;
	using pointer = Pointer;
	using const_pointer = ConstPointer;
	using iterator = Iterator;
	using const_iterator = ConstIterator;
	using reverse_iterator = ReverseIterator;
	using const_reverse_iterator = ConstReverseIterator;
	using difference_type = Difference;

public:
	static constexpr Size size() noexcept { return dim; }
	static constexpr Size length() noexcept { return dim; }
	static constexpr Size max_size() noexcept { return dim; }
	static constexpr bool empty() noexcept { return dim == 0; }

public:
	pointer data_[dim];

public:
	template<typename... Args, typename = typename
		std::enable_if_t<
			std::is_convertible<
				std::tuple<Args...>,
				TypeTuple<value_type, dim>
			>::value
		>>
	constexpr Vec(Args&&... args) : data_{&args...} {}
	~Vec() noexcept = default;

	constexpr Vec(const RefVecType& other) noexcept = default;
	constexpr Vec(const VecType& other) noexcept
		{ for(Size i(0); i < dim; ++i) data_[i] = &other[i]; }

	constexpr RefVecType& operator=(const VecType& other) noexcept
		{ for(Size i(0); i < dim; ++i) *data_[i] = other[i]; return *this; }

	//operator
	template<Size OD, typename OT> constexpr VecType& operator+=(const Vec<OD, OT>& lhs)
		{ for(Size i = 0; i < min(lhs.size(), dim); i++) (*this)[i] += lhs[i];  return *this; }
	template<Size OD, typename OT> constexpr VecType& operator-=(const Vec<OD, OT>& lhs)
		{ for(Size i = 0; i < min(lhs.size(), dim); i++) (*this)[i] -= lhs[i];  return *this; }
	template<Size OD, typename OT> constexpr VecType& operator*=(const Vec<OD, OT>& lhs)
		{ for(Size i = 0; i < min(lhs.size(), dim); i++) (*this)[i] *= lhs[i];  return *this; }
	template<Size OD, typename OT> constexpr VecType& operator/=(const Vec<OD, OT>& lhs)
		{ for(Size i = 0; i < min(lhs.size(), dim); i++) (*this)[i] /= lhs[i];  return *this; }
	template<Size OD, typename OT> constexpr VecType& operator%=(const Vec<OD, OT>& lhs)
		{ for(Size i = 0; i < min(lhs.size(), dim); i++) (*this)[i] %= lhs[i];  return *this; }
	template<Size OD, typename OT> constexpr VecType& operator|=(const Vec<OD, OT>& lhs)
		{ for(Size i = 0; i < min(lhs.size(), dim); i++) (*this)[i] |= lhs[i];  return *this; }
	template<Size OD, typename OT> constexpr VecType& operator^=(const Vec<OD, OT>& lhs)
		{ for(Size i = 0; i < min(lhs.size(), dim); i++) (*this)[i] ^= lhs[i];  return *this; }
	template<Size OD, typename OT> constexpr VecType& operator&=(const Vec<OD, OT>& lhs)
		{ for(Size i = 0; i < min(lhs.size(), dim); i++) (*this)[i] &= lhs[i];  return *this; }
	template<Size OD, typename OT> constexpr VecType& operator>>=(const Vec<OD, OT>& lhs)
		{ for(size_t i = 0; i < min(lhs.size(), dim); i++) (*this)[i] >>= lhs[i]; return *this; }
	template<Size OD, typename OT> constexpr VecType& operator<<=(const Vec<OD, OT>& lhs)
		{ for(size_t i = 0; i < min(lhs.size(), dim); i++) (*this)[i] <<= lhs[i]; return *this; }

	template<typename OT> constexpr VecType& operator+=(OT lhs)
		{ for(auto& val : *this) val += lhs; return *this; }
	template<typename OT> constexpr VecType& operator-=(OT lhs)
		{ for(auto& val : *this) val -= lhs; return *this; }
	template<typename OT> constexpr VecType& operator*=(OT lhs)
		{ for(auto& val : *this) val *= lhs; return *this; }
	template<typename OT> constexpr VecType& operator/=(OT lhs)
		{ for(auto& val : *this) val /= lhs; return *this; }
	template<typename OT> constexpr VecType& operator%=(OT lhs)
		{ for(auto& val : *this) val %= lhs; return *this; }
	template<typename OT> constexpr VecType& operator|=(OT lhs)
		{ for(auto& val : *this) val |= lhs; return *this; }
	template<typename OT> constexpr VecType& operator^=(OT lhs)
		{ for(auto& val : *this) val ^= lhs; return *this; }
	template<typename OT> constexpr VecType& operator&=(OT lhs)
		{ for(auto& val : *this) val &= lhs; return *this; }
	template<typename OT> constexpr VecType& operator>>=(OT lhs)
		{ for(auto& val : *this) val >>= lhs; return *this; }
	template<typename OT> constexpr VecType& operator<<=(OT lhs)
		{ for(auto& val : *this) val <<= lhs; return *this; }

	constexpr VecType operator-() const
		{ VecType ret{}; for(size_t i(0); i < size(); i++) ret[i] -= (*this)[i]; return ret; }

	template <Size OD, typename OT, typename =
		typename std::enable_if<!std::is_reference<OT>::value>::type>
	constexpr operator Vec<OD, OT>() const
	{
		auto ret = Vec<OD, OT> (size());
		for(size_t i(0); i < min(size(), ret.size()); i++)
			ret[i] = (*this)[i];
		return ret;
	}

	template<Size S>
	constexpr Vec<S, T> subVec(Size position = 0) const
	{
		auto ret = Vec<S, T> {};
		for(std::size_t i(0); i < min(S, size() - position); ++i)
			ret[i] = (*this)[position + i];

		return ret;
	}

	Vec<dynamicSize, T> subVec(Size position = 0, Size psize = -1) const
	{
		auto ret = Vec<dynamicSize, T>(size);
		for(std::size_t i(0); i < min(psize, size() - position); ++i)
			ret[i] = (*this)[position + i];

		return ret;
	}

	//stl
	//TODO: correct data implementation
	//could be done holding an internal Vec. Data would become invalid and must be const...
	//just not implement data?
	//ConstPointer data() const noexcept { return data_; }
	//pointer data() noexcept { return data_; }

	constexpr void fill(const Value& val) { for(auto& v : data_) *v = val; }

	constexpr Iterator begin() noexcept { return Iterator(*this); }
	constexpr ConstIterator begin() const noexcept { return ConstIterator(*this); }
	constexpr ConstIterator cbegin() const noexcept { return ConstIterator(*this); }

	constexpr Iterator end() noexcept { return Iterator(*this, dim); }
	constexpr ConstIterator end() const noexcept { return ConstIterator(*this, dim); }
	constexpr ConstIterator cend() const noexcept { return ConstIterator(*this, dim); }

	constexpr ReverseIterator rbegin() noexcept { return ReverseIterator(end()); }
	constexpr ConstReverseIterator rbegin() const noexcept { return ConstReverseIterator(cend()); }
	constexpr ConstReverseIterator crbegin() const noexcept { return ConstReverseIterator(cend()); }

	constexpr ReverseIterator rend() noexcept { return ReverseIterator(begin()); }
	constexpr ConstReverseIterator rend() const noexcept { return ConstReverseIterator(cbegin()); }
	constexpr ConstReverseIterator crend() const noexcept { return constReverseIterator(cbegin()); }

	constexpr Reference operator[](Size i){ return *data_[i]; }
	constexpr ConstReference operator[](Size i) const { return *data_[i]; }

	constexpr Reference at(Size i)
		{ if(i >= dim || i < 0) throw RangeError("nytl::Vec::at"); return *data_[i]; }
	constexpr ConstReference at(Size i) const
		{ if(i >= dim || i < 0) throw RangeError("nytl::Vec::at"); return *data_[i]; }

	constexpr Reference front() noexcept { return *data_[0]; }
	constexpr ConstReference front() const noexcept { return *data_[0]; }

	constexpr Reference back() noexcept { return *data_[dim - 1]; }
	constexpr ConstReference back() const noexcept { return *data_[dim - 1]; }

	constexpr void swap(VecType& other)
		{ for(Size i(0); i < size(); ++i) std::swap(*data_[i], other[i]); }

	friend constexpr void swap(RefVecType& a, RefVecType& b) { a.swap(b); }
};

template<typename T> class Vec<2, T&>
{
public:
	static constexpr std::size_t dim = 2;

	using Size = std::size_t;
	using Precision = T;
	using Value = T;
	using VecType = Vec<dim, T>;
	using RefVecType = Vec<dim, T&>;

	using Reference = Value&;
	using ConstReference = const Value&;
	using Pointer = Value*;
	using ConstPointer = const Value*;
	using Iterator = RefVecIterator<dim, Value>;
	using ConstIterator = ConstRefVecIterator<dim, Value>;
	using ReverseIterator = std::reverse_iterator<Iterator>;
	using ConstReverseIterator = std::reverse_iterator<ConstIterator>;
	using Difference = Size;

	using RangeError = std::out_of_range;

	//stl
	using size_type = Size;
	using value_type = Value;
	using reference = Reference;
	using const_reference = ConstReference;
	using pointer = Pointer;
	using const_pointer = ConstPointer;
	using iterator = Iterator;
	using const_iterator = ConstIterator;
	using reverse_iterator = ReverseIterator;
	using const_reverse_iterator = ConstReverseIterator;
	using difference_type = Difference;

public:
	static constexpr Size size() noexcept { return dim; }
	static constexpr Size length() noexcept { return dim; }
	static constexpr Size max_size() noexcept { return dim; }
	static constexpr bool empty() noexcept { return dim == 0; }

public:
	Reference x;
	Reference y;

public:
	constexpr Vec(Reference a, Reference b) : x(a), y(b) {}
	~Vec() noexcept = default;

	constexpr Vec(const RefVecType& other) noexcept = default;

	constexpr Vec(VecType& other) noexcept : x(other.x), y(other.y) {}
	constexpr RefVecType& operator=(const VecType& other) noexcept
		{ x = other.x; y = other.y; return *this; }

	//operator
	template<Size OD, typename OT> constexpr VecType& operator+=(const Vec<OD, OT>& lhs)
		{ for(Size i = 0; i < min(lhs.size(), dim); i++) (*this)[i] += lhs[i];  return *this; }
	template<Size OD, typename OT> constexpr VecType& operator-=(const Vec<OD, OT>& lhs)
		{ for(Size i = 0; i < min(lhs.size(), dim); i++) (*this)[i] -= lhs[i];  return *this; }
	template<Size OD, typename OT> constexpr VecType& operator*=(const Vec<OD, OT>& lhs)
		{ for(Size i = 0; i < min(lhs.size(), dim); i++) (*this)[i] *= lhs[i];  return *this; }
	template<Size OD, typename OT> constexpr VecType& operator/=(const Vec<OD, OT>& lhs)
		{ for(Size i = 0; i < min(lhs.size(), dim); i++) (*this)[i] /= lhs[i];  return *this; }
	template<Size OD, typename OT> constexpr VecType& operator%=(const Vec<OD, OT>& lhs)
		{ for(Size i = 0; i < min(lhs.size(), dim); i++) (*this)[i] %= lhs[i];  return *this; }
	template<Size OD, typename OT> constexpr VecType& operator|=(const Vec<OD, OT>& lhs)
		{ for(Size i = 0; i < min(lhs.size(), dim); i++) (*this)[i] |= lhs[i];  return *this; }
	template<Size OD, typename OT> constexpr VecType& operator^=(const Vec<OD, OT>& lhs)
		{ for(Size i = 0; i < min(lhs.size(), dim); i++) (*this)[i] ^= lhs[i];  return *this; }
	template<Size OD, typename OT> constexpr VecType& operator&=(const Vec<OD, OT>& lhs)
		{ for(Size i = 0; i < min(lhs.size(), dim); i++) (*this)[i] &= lhs[i];  return *this; }
	template<Size OD, typename OT> constexpr VecType& operator>>=(const Vec<OD, OT>& lhs)
		{ for(size_t i = 0; i < min(lhs.size(), dim); i++) (*this)[i] >>= lhs[i]; return *this; }
	template<Size OD, typename OT> constexpr VecType& operator<<=(const Vec<OD, OT>& lhs)
		{ for(size_t i = 0; i < min(lhs.size(), dim); i++) (*this)[i] <<= lhs[i]; return *this; }

	template<typename OT> constexpr VecType& operator+=(OT lhs)
		{ for(auto& val : *this) val += lhs; return *this; }
	template<typename OT> constexpr VecType& operator-=(OT lhs)
		{ for(auto& val : *this) val -= lhs; return *this; }
	template<typename OT> constexpr VecType& operator*=(OT lhs)
		{ for(auto& val : *this) val *= lhs; return *this; }
	template<typename OT> constexpr VecType& operator/=(OT lhs)
		{ for(auto& val : *this) val /= lhs; return *this; }
	template<typename OT> constexpr VecType& operator%=(OT lhs)
		{ for(auto& val : *this) val %= lhs; return *this; }
	template<typename OT> constexpr VecType& operator|=(OT lhs)
		{ for(auto& val : *this) val |= lhs; return *this; }
	template<typename OT> constexpr VecType& operator^=(OT lhs)
		{ for(auto& val : *this) val ^= lhs; return *this; }
	template<typename OT> constexpr VecType& operator&=(OT lhs)
		{ for(auto& val : *this) val &= lhs; return *this; }
	template<typename OT> constexpr VecType& operator>>=(OT lhs)
		{ for(auto& val : *this) val >>= lhs; return *this; }
	template<typename OT> constexpr VecType& operator<<=(OT lhs)
		{ for(auto& val : *this) val <<= lhs; return *this; }

	constexpr VecType operator-() const { return VecType(-x, -y); }

	template <Size OD, typename OT, typename =
		typename std::enable_if<!std::is_reference<OT>::value>::type>
	constexpr operator Vec<OD, OT>() const
	{
		auto ret = Vec<OD, OT> (size());
		for(size_t i(0); i < min(size(), ret.size()); i++)
			ret[i] = (*this)[i];
		return ret;
	}

	template<Size S>
	constexpr Vec<S, T> subVec(Size position = 0) const
	{
		auto ret = Vec<S, T> {};
		for(std::size_t i(0); i < min(S, size() - position); ++i)
			ret[i] = (*this)[position + i];

		return ret;
	}

	Vec<dynamicSize, T> subVec(Size position = 0, Size psize = -1) const
	{
		auto ret = Vec<dynamicSize, T>(size);
		for(std::size_t i(0); i < min(psize, size() - position); ++i)
			ret[i] = (*this)[position + i];

		return ret;
	}

	//stl norm stuff, std::array
	constexpr ConstPointer data() const noexcept { return &x; }
	constexpr Pointer data() noexcept { return &x; }

	constexpr void fill(const Value& val) { x = val; y = val; }

	constexpr Iterator begin() noexcept { return Iterator(*this); }
	constexpr ConstIterator begin() const noexcept { return ConstIterator(*this); }
	constexpr ConstIterator cbegin() const noexcept { return ConstIterator(*this); }

	constexpr Iterator end() noexcept { return Iterator(*this, dim); }
	constexpr ConstIterator end() const noexcept { return ConstIterator(*this, dim); }
	constexpr ConstIterator cend() const noexcept { return ConstIterator(*this, dim); }

	constexpr ReverseIterator rbegin() noexcept { return ReverseIterator(end()); }
	constexpr ConstReverseIterator rbegin() const noexcept { return ConstReverseIterator(cend()); }
	constexpr ConstReverseIterator crbegin() const noexcept { return ConstReverseIterator(cend()); }

	constexpr ReverseIterator rend() noexcept { return ReverseIterator(begin()); }
	constexpr ConstReverseIterator rend() const noexcept { return ConstReverseIterator(cbegin()); }
	constexpr ConstReverseIterator crend() const noexcept { return ConstReverseIterator(cbegin()); }

	constexpr Reference operator[](Size i){ if(i == 0) return x; return y; }
	constexpr ConstReference operator[](Size i) const { if(i == 0) return x; return y; }

	constexpr Reference at(Size i)
		{ if(i >= dim)	throw RangeError("nytl::RefVec::at"); return (*this)[i]; }
	constexpr ConstReference at(Size i) const
		{ if(i >= dim) throw std::out_of_range("nytl::RefVec::at"); return (*this)[i]; }

	constexpr Reference front() noexcept { return x; }
	constexpr ConstReference front() const noexcept { return x; }

	constexpr Reference back() noexcept { return y; }
	constexpr ConstReference back() const noexcept { return y; }

	constexpr void swap(VecType& other){ std::swap(x, other.x); std::swap(y, other.y); }
	friend constexpr void swap(RefVecType& a, RefVecType& b) { a.swap(b); }
};

template<typename T> class Vec<3, T&>
{
public:
	static constexpr std::size_t dim = 3;

	using Size = std::size_t;
	using Precision = T;
	using Value = T;
	using VecType = Vec<dim, T>;
	using RefVecType = Vec<dim, T&>;

	using Reference = Value&;
	using ConstReference = const Value&;
	using Pointer = Value*;
	using ConstPointer = const Value*;
	using Iterator = RefVecIterator<dim, Value>;
	using ConstIterator = ConstRefVecIterator<dim, Value>;
	using ReverseIterator = std::reverse_iterator<Iterator>;
	using ConstReverseIterator = std::reverse_iterator<ConstIterator>;
	using Difference = Size;

	using RangeError = std::out_of_range;

	//stl
	using size_type = Size;
	using value_type = Value;
	using reference = Reference;
	using const_reference = ConstReference;
	using pointer = Pointer;
	using const_pointer = ConstPointer;
	using iterator = Iterator;
	using const_iterator = ConstIterator;
	using reverse_iterator = ReverseIterator;
	using const_reverse_iterator = ConstReverseIterator;
	using difference_type = Difference;

public:
	static constexpr Size size() noexcept { return dim; }
	static constexpr Size length() noexcept { return dim; }
	static constexpr Size max_size() noexcept { return dim; }
	static constexpr bool empty() noexcept { return dim == 0; }

public:
	Reference x;
	Reference y;
	Reference z;

public:
	constexpr Vec(Reference a, Reference b, Reference c) noexcept : x(a), y(b), z(c) {}
	~Vec() noexcept = default;

	constexpr Vec(const RefVecType& other) noexcept = default;

	constexpr Vec(const VecType& other) noexcept : x(other.x), y(other.y), z(other.z) {}
	constexpr RefVecType& operator=(const VecType& other) noexcept
		{ x = other.x; y = other.y; z = other.z; return *this; }

	//operator
	template<Size OD, typename OT> constexpr VecType& operator+=(const Vec<OD, OT>& lhs)
		{ for(Size i = 0; i < min(lhs.size(), dim); i++) (*this)[i] += lhs[i];  return *this; }
	template<Size OD, typename OT> constexpr VecType& operator-=(const Vec<OD, OT>& lhs)
		{ for(Size i = 0; i < min(lhs.size(), dim); i++) (*this)[i] -= lhs[i];  return *this; }
	template<Size OD, typename OT> constexpr VecType& operator*=(const Vec<OD, OT>& lhs)
		{ for(Size i = 0; i < min(lhs.size(), dim); i++) (*this)[i] *= lhs[i];  return *this; }
	template<Size OD, typename OT> constexpr VecType& operator/=(const Vec<OD, OT>& lhs)
		{ for(Size i = 0; i < min(lhs.size(), dim); i++) (*this)[i] /= lhs[i];  return *this; }
	template<Size OD, typename OT> constexpr VecType& operator%=(const Vec<OD, OT>& lhs)
		{ for(Size i = 0; i < min(lhs.size(), dim); i++) (*this)[i] %= lhs[i];  return *this; }
	template<Size OD, typename OT> constexpr VecType& operator|=(const Vec<OD, OT>& lhs)
		{ for(Size i = 0; i < min(lhs.size(), dim); i++) (*this)[i] |= lhs[i];  return *this; }
	template<Size OD, typename OT> constexpr VecType& operator^=(const Vec<OD, OT>& lhs)
		{ for(Size i = 0; i < min(lhs.size(), dim); i++) (*this)[i] ^= lhs[i];  return *this; }
	template<Size OD, typename OT> constexpr VecType& operator&=(const Vec<OD, OT>& lhs)
		{ for(Size i = 0; i < min(lhs.size(), dim); i++) (*this)[i] &= lhs[i];  return *this; }
	template<Size OD, typename OT> constexpr VecType& operator>>=(const Vec<OD, OT>& lhs)
		{ for(size_t i = 0; i < min(lhs.size(), dim); i++) (*this)[i] >>= lhs[i]; return *this; }
	template<Size OD, typename OT> constexpr VecType& operator<<=(const Vec<OD, OT>& lhs)
		{ for(size_t i = 0; i < min(lhs.size(), dim); i++) (*this)[i] <<= lhs[i]; return *this; }

	template<typename OT> constexpr VecType& operator+=(OT lhs)
		{ for(auto& val : *this) val += lhs; return *this; }
	template<typename OT> constexpr VecType& operator-=(OT lhs)
		{ for(auto& val : *this) val -= lhs; return *this; }
	template<typename OT> constexpr VecType& operator*=(OT lhs)
		{ for(auto& val : *this) val *= lhs; return *this; }
	template<typename OT> constexpr VecType& operator/=(OT lhs)
		{ for(auto& val : *this) val /= lhs; return *this; }
	template<typename OT> constexpr VecType& operator%=(OT lhs)
		{ for(auto& val : *this) val %= lhs; return *this; }
	template<typename OT> constexpr VecType& operator|=(OT lhs)
		{ for(auto& val : *this) val |= lhs; return *this; }
	template<typename OT> constexpr VecType& operator^=(OT lhs)
		{ for(auto& val : *this) val ^= lhs; return *this; }
	template<typename OT> constexpr VecType& operator&=(OT lhs)
		{ for(auto& val : *this) val &= lhs; return *this; }
	template<typename OT> constexpr VecType& operator>>=(OT lhs)
		{ for(auto& val : *this) val >>= lhs; return *this; }
	template<typename OT> constexpr VecType& operator<<=(OT lhs)
		{ for(auto& val : *this) val <<= lhs; return *this; }

	constexpr VecType operator-() const { return VecType(-x, -y, -z); }

	template <Size OD, typename OT, typename =
		typename std::enable_if<!std::is_reference<OT>::value>::type>
	constexpr operator Vec<OD, OT>() const
	{
		auto ret = Vec<OD, OT> (size());
		for(size_t i(0); i < min(size(), ret.size()); i++)
			ret[i] = (*this)[i];
		return ret;
	}

	template<Size S>
	constexpr Vec<S, T> subVec(Size position = 0) const
	{
		auto ret = Vec<S, T> {};
		for(std::size_t i(0); i < min(S, size() - position); ++i)
			ret[i] = (*this)[position + i];

		return ret;
	}

	Vec<dynamicSize, T> subVec(Size position = 0, Size psize = -1) const
	{
		auto ret = Vec<dynamicSize, T>(size);
		for(std::size_t i(0); i < min(psize, size() - position); ++i)
			ret[i] = (*this)[position + i];

		return ret;
	}

	//stl norm stuff, std::array
	constexpr ConstPointer data() const noexcept { return &x; }
	constexpr Pointer data() noexcept { return &x; }

	constexpr void fill(const Value& val) { x = val; y = val; z = val; }

	constexpr Iterator begin() noexcept { return Iterator(*this); }
	constexpr ConstIterator begin() const noexcept { return ConstIterator(*this); }
	constexpr ConstIterator cbegin() const noexcept { return ConstIterator(*this); }

	constexpr Iterator end() noexcept { return Iterator(*this, dim); }
	constexpr ConstIterator end() const noexcept { return ConstIterator(*this, dim); }
	constexpr ConstIterator cend() const noexcept { return ConstIterator(*this, dim); }

	constexpr ReverseIterator rbegin() noexcept { return ReverseIterator(end()); }
	constexpr ConstReverseIterator rbegin() const noexcept { return ConstReverseIterator(cend()); }
	constexpr ConstReverseIterator crbegin() const noexcept { return ConstReverseIterator(cend()); }

	constexpr ReverseIterator rend() noexcept { return ReverseIterator(begin()); }
	constexpr ConstReverseIterator rend() const noexcept { return ConstReverseIterator(cbegin()); }
	constexpr ConstReverseIterator crend() const noexcept { return ConstReverseIterator(cbegin()); }

	constexpr Reference operator[](Size i){ if(i == 0) return x; if(i == 1) return y; return z; }
	constexpr ConstReference operator[](Size i) const { if(i == 0) return x; if(i == 1) return y; return z; }

	constexpr Reference at(Size i)
		{ if(i >= dim)	throw RangeError("nytl::RefVec::at"); return (*this)[i]; }
	constexpr ConstReference at(Size i) const
		{ if(i >= dim) throw RangeError("nytl::RefVec::at"); return (*this)[i]; }

	constexpr Reference front() noexcept { return x; }
	constexpr ConstReference front() const noexcept { return x; }

	constexpr Reference back() noexcept { return z; }
	constexpr ConstReference back() const noexcept { return z; }

	constexpr void swap(VecType& other)
		{ std::swap(x, other.x); std::swap(y, other.y); std::swap(z, other.z); }
	friend constexpr void swap(RefVecType& a, RefVecType& b) { a.swap(b); }

	//custom
	constexpr Vec2<Reference> xy() const noexcept { return Vec2<Reference>(x,y); }
	constexpr Vec2<Reference> yz() const noexcept { return Vec2<Reference>(y,z); }
	constexpr Vec2<Reference> xz() const noexcept { return Vec2<Reference>(x,z); }
};

template<typename T> class Vec<4, T&>
{
public:
	static constexpr std::size_t dim = 4;

	using Size = std::size_t;
	using Precision = T;
	using Value = T;
	using VecType = Vec<dim, T>;
	using RefVecType = Vec<dim, T&>;

	using Reference = Value&;
	using ConstReference = const Value&;
	using Pointer = Value*;
	using ConstPointer = const Value*;
	using Iterator = RefVecIterator<dim, Value>;
	using ConstIterator = ConstRefVecIterator<dim, Value>;
	using ReverseIterator = std::reverse_iterator<Iterator>;
	using ConstReverseIterator = std::reverse_iterator<ConstIterator>;
	using Difference = Size;

	using RangeError = std::out_of_range;

	//stl
	using size_type = Size;
	using value_type = Value;
	using reference = Reference;
	using const_reference = ConstReference;
	using pointer = Pointer;
	using const_pointer = ConstPointer;
	using iterator = Iterator;
	using const_iterator = ConstIterator;
	using reverse_iterator = ReverseIterator;
	using const_reverse_iterator = ConstReverseIterator;
	using difference_type = Difference;

public:
	static constexpr Size size() noexcept { return dim; }
	static constexpr Size length() noexcept { return dim; }
	static constexpr Size max_size() noexcept { return dim; }
	static constexpr bool empty() noexcept { return dim == 0; }

public:
	Reference x;
	Reference y;
	Reference z;
	Reference w;

public:
	constexpr Vec(Reference a, Reference b, Reference c, Reference d)
		noexcept : x(a), y(b), z(c), w(d) {}
	~Vec() noexcept = default;

	constexpr Vec(const RefVecType& other) noexcept = default;

	constexpr Vec(const VecType& other) noexcept : x(other.x), y(other.y), z(other.z), w(other.w) {}
	constexpr RefVecType& operator=(const VecType& other) noexcept
		{ x = other.x; y = other.y; z = other.z; w = other.w; return *this; }

	//operator
	template<Size OD, typename OT> constexpr VecType& operator+=(const Vec<OD, OT>& lhs)
		{ for(Size i = 0; i < min(lhs.size(), dim); i++) (*this)[i] += lhs[i];  return *this; }
	template<Size OD, typename OT> constexpr VecType& operator-=(const Vec<OD, OT>& lhs)
		{ for(Size i = 0; i < min(lhs.size(), dim); i++) (*this)[i] -= lhs[i];  return *this; }
	template<Size OD, typename OT> constexpr VecType& operator*=(const Vec<OD, OT>& lhs)
		{ for(Size i = 0; i < min(lhs.size(), dim); i++) (*this)[i] *= lhs[i];  return *this; }
	template<Size OD, typename OT> constexpr VecType& operator/=(const Vec<OD, OT>& lhs)
		{ for(Size i = 0; i < min(lhs.size(), dim); i++) (*this)[i] /= lhs[i];  return *this; }
	template<Size OD, typename OT> constexpr VecType& operator%=(const Vec<OD, OT>& lhs)
		{ for(Size i = 0; i < min(lhs.size(), dim); i++) (*this)[i] %= lhs[i];  return *this; }
	template<Size OD, typename OT> constexpr VecType& operator|=(const Vec<OD, OT>& lhs)
		{ for(Size i = 0; i < min(lhs.size(), dim); i++) (*this)[i] |= lhs[i];  return *this; }
	template<Size OD, typename OT> constexpr VecType& operator^=(const Vec<OD, OT>& lhs)
		{ for(Size i = 0; i < min(lhs.size(), dim); i++) (*this)[i] ^= lhs[i];  return *this; }
	template<Size OD, typename OT> constexpr VecType& operator&=(const Vec<OD, OT>& lhs)
		{ for(Size i = 0; i < min(lhs.size(), dim); i++) (*this)[i] &= lhs[i];  return *this; }
	template<Size OD, typename OT> constexpr VecType& operator>>=(const Vec<OD, OT>& lhs)
		{ for(size_t i = 0; i < min(lhs.size(), dim); i++) (*this)[i] >>= lhs[i]; return *this; }
	template<Size OD, typename OT> constexpr VecType& operator<<=(const Vec<OD, OT>& lhs)
		{ for(size_t i = 0; i < min(lhs.size(), dim); i++) (*this)[i] <<= lhs[i]; return *this; }

	template<typename OT> constexpr VecType& operator+=(OT lhs)
		{ for(auto& val : *this) val += lhs; return *this; }
	template<typename OT> constexpr VecType& operator-=(OT lhs)
		{ for(auto& val : *this) val -= lhs; return *this; }
	template<typename OT> constexpr VecType& operator*=(OT lhs)
		{ for(auto& val : *this) val *= lhs; return *this; }
	template<typename OT> constexpr VecType& operator/=(OT lhs)
		{ for(auto& val : *this) val /= lhs; return *this; }
	template<typename OT> constexpr VecType& operator%=(OT lhs)
		{ for(auto& val : *this) val %= lhs; return *this; }
	template<typename OT> constexpr VecType& operator|=(OT lhs)
		{ for(auto& val : *this) val |= lhs; return *this; }
	template<typename OT> constexpr VecType& operator^=(OT lhs)
		{ for(auto& val : *this) val ^= lhs; return *this; }
	template<typename OT> constexpr VecType& operator&=(OT lhs)
		{ for(auto& val : *this) val &= lhs; return *this; }
	template<typename OT> constexpr VecType& operator>>=(OT lhs)
		{ for(auto& val : *this) val >>= lhs; return *this; }
	template<typename OT> constexpr VecType& operator<<=(OT lhs)
		{ for(auto& val : *this) val <<= lhs; return *this; }

	constexpr VecType operator-() const { return VecType(x, y, z); }

	template <Size OD, typename OT, typename =
		typename std::enable_if<!std::is_reference<OT>::value>::type>
	constexpr operator Vec<OD, OT>() const
	{
		auto ret = Vec<OD, OT> (size());
		for(size_t i(0); i < min(size(), ret.size()); i++)
			ret[i] = (*this)[i];
		return ret;
	}

	template<Size S>
	constexpr Vec<S, T> subVec(Size position = 0) const
	{
		auto ret = Vec<S, T> {};
		for(std::size_t i(0); i < min(S, size() - position); ++i)
			ret[i] = (*this)[position + i];

		return ret;
	}

	Vec<dynamicSize, T> subVec(Size position = 0, Size psize = -1) const
	{
		auto ret = Vec<dynamicSize, T>(size);
		for(std::size_t i(0); i < min(psize, size() - position); ++i)
			ret[i] = (*this)[position + i];

		return ret;
	}

	//stl norm stuff, std::array
	constexpr ConstPointer data() const noexcept { return &x; }
	constexpr Pointer data() noexcept { return &x; }

	constexpr void fill(const Value& val) { x = val; y = val; z = val; w = val; }

	constexpr Iterator begin() noexcept { return Iterator(*this); }
	constexpr ConstIterator begin() const noexcept { return ConstIterator(*this); }
	constexpr ConstIterator cbegin() const noexcept { return ConstIterator(*this); }

	constexpr Iterator end() noexcept { return Iterator(*this, dim); }
	constexpr ConstIterator end() const noexcept { return ConstIterator(*this, dim); }
	constexpr ConstIterator cend() const noexcept { return ConstIterator(*this, dim); }

	constexpr ReverseIterator rbegin() noexcept { return ReverseIterator(end()); }
	constexpr ConstReverseIterator rbegin() const noexcept { return ConstReverseIterator(cend()); }
	constexpr ConstReverseIterator crbegin() const noexcept { return ConstReverseIterator(cend()); }

	constexpr ReverseIterator rend() noexcept { return ReverseIterator(begin()); }
	constexpr ConstReverseIterator rend() const noexcept { return ConstReverseIterator(cbegin()); }
	constexpr ConstReverseIterator crend() const noexcept { return ConstReverseIterator(cbegin()); }

	constexpr Reference operator[](Size i)
		{ if(i == 0) return x; if(i == 1) return y; if(i == 2) return z; return w; }
	constexpr ConstReference operator[](Size i) const
		{ if(i == 0) return x; if(i == 1) return y; if(i == 2) return z; return w; }

	constexpr Reference at(Size i)
		{ if(i >= dim)	throw RangeError("nytl::RefVec::at"); return (*this)[i]; }
	constexpr ConstReference at(Size i) const
		{ if(i >= dim) throw RangeError("nytl::RefVec::at"); return (*this)[i]; }

	constexpr Reference front() noexcept { return x; }
	constexpr ConstReference front() const noexcept { return x; }

	constexpr Reference back() noexcept { return w; }
	constexpr ConstReference back() const noexcept { return w; }

	constexpr void swap(VecType& other)
		{ using std::swap; swap(x, other.x); swap(y, other.y); swap(z, other.z); swap(w, other.w);}

	friend constexpr void swap(RefVecType& a, RefVecType& b){ a.swap(b); }

	//custom
	constexpr Vec2<Reference> xy() const noexcept { return Vec2<Reference>(x,y); }
	constexpr Vec2<Reference> xz() const noexcept { return Vec2<Reference>(x,z); }
	constexpr Vec2<Reference> xw() const noexcept { return Vec2<Reference>(x,w); }
	constexpr Vec2<Reference> yz() const noexcept { return Vec2<Reference>(y,z); }
	constexpr Vec2<Reference> yw() const noexcept { return Vec2<Reference>(y,w); }
	constexpr Vec2<Reference> zw() const noexcept { return Vec2<Reference>(z,w); }

	constexpr Vec3<Reference> xyz() const noexcept { return Vec3<Reference>(x,y,z); }
	constexpr Vec3<Reference> xyw() const noexcept { return Vec3<Reference>(x,y,w); }
	constexpr Vec3<Reference> xzw() const noexcept { return Vec3<Reference>(x,z,w); }
	constexpr Vec3<Reference> yzw() const noexcept { return Vec3<Reference>(y,z,w); }
};

template<size_t dim, typename T> constexpr size_t Vec<dim, T&>::dim;
template<typename T> constexpr size_t Vec<2, T&>::dim;
template<typename T> constexpr size_t Vec<3, T&>::dim;
template<typename T> constexpr size_t Vec<4, T&>::dim;

}

#endif //header guard
